#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <boost/scoped_array.hpp>
//      PE-
#include <pe_32_64.h>
#include <pe_resource_manager.h>
//   LZO1Z999
#include "../../lzo-2.06/include/lzo/lzo1z.h"
//    
#include "structs.h"
//    
#include "../unpacker/parameters.h"
//  ()
#include "unpacker.h"

//      PE  LZO
#ifndef _M_X64
#ifdef _DEBUG
#pragma comment(lib, "../../Debug/pe_lib.lib")
#pragma comment(lib, "../Debug/lzo-2.06.lib")
#else
#pragma comment(lib, "../../Release/pe_lib.lib")
#pragma comment(lib, "../Release/lzo-2.06.lib")
#endif
#else
#ifdef _DEBUG
#pragma comment(lib, "../../x64/Debug/pe_lib.lib")
#pragma comment(lib, "../x64/Debug/lzo-2.06.lib")
#else
#pragma comment(lib, "../../x64/Release/pe_lib.lib")
#pragma comment(lib, "../x64/Release/lzo-2.06.lib")
#endif
#endif

int main(int argc, char* argv[])
{
	// ,    
	//       , 
	//   ,    
	// ,    
	if(argc != 2)
	{
		std::cout << "Usage: simple_pe_packer.exe PE_FILE" << std::endl;
		return 0;
	}

	//  -      argv   1
	std::ifstream file(argv[1], std::ios::in | std::ios::binary);
	if(!file)
	{
		//     -     
		std::cout << "Cannot open " << argv[1] << std::endl;
		return -1;
	}

	try
	{
		//    32- PE-
		//   false,     
		//""      
		//""   
		//    ,     
		pe32 image(file, false, false);

		//,  .NET    
		if(image.is_dotnet())
		{
			std::cout << ".NEt image cannot be packed!" << std::endl;
			return -1;
		}

		//   ,  ,    
		{
			std::cout << "Entropy of sections: ";
			double entropy = image.calculate_entropy();
			std::cout << entropy << std::endl;
			// wasm.ru  ,   ,
			//  PE-    6.8
			// ,  ,  , 
			// ( )    
			//  ,    
			if(entropy > 6.8)
			{
				std::cout << "File has already been packed!" << std::endl;
				//return -1;
			}
		}

		//   LZO
		if(lzo_init() != LZO_E_OK)
		{
			std::cout << "Error initializing LZO library" << std::endl;
			return -1;
		}
		
		std::cout << "Reading sections..." << std::endl;

		//   PE-
		const pe_base::section_list& sections = image.get_image_sections();
		if(sections.empty())
		{
			//      ,   
			std::cout << "File has no sections!" << std::endl;
			return -1;
		}

		//    PE-
		packed_file_info basic_info = {0};
		//     
		basic_info.number_of_sections = sections.size();
		//   LOCK
		basic_info.lock_opcode = 0xf0;

		//    
		//    
		basic_info.original_import_directory_rva = image.get_directory_rva(IMAGE_DIRECTORY_ENTRY_IMPORT);
		basic_info.original_import_directory_size = image.get_directory_size(IMAGE_DIRECTORY_ENTRY_IMPORT);
		//    
		//    
		basic_info.original_resource_directory_rva = image.get_directory_rva(IMAGE_DIRECTORY_ENTRY_RESOURCE);
		basic_info.original_resource_directory_size = image.get_directory_size(IMAGE_DIRECTORY_ENTRY_RESOURCE);
		//    
		//    
		basic_info.original_relocation_directory_rva = image.get_directory_rva(IMAGE_DIRECTORY_ENTRY_BASERELOC);
		basic_info.original_relocation_directory_size = image.get_directory_size(IMAGE_DIRECTORY_ENTRY_BASERELOC);
		
		//  
		//     
		basic_info.original_load_config_directory_rva = image.get_directory_rva(IMAGE_DIRECTORY_ENTRY_LOAD_CONFIG);

		//   
		basic_info.original_entry_point = image.get_ep();
		//     
		// 
		basic_info.total_virtual_size_of_sections = image.get_size_of_image();


		//,    
		// packed_section   
		std::string packed_sections_info;

		{
			//        
			packed_sections_info.resize(sections.size() * sizeof(packed_section));

			//""   ,      
			std::string raw_section_data;
			//  
			unsigned long current_section = 0;

			//  
			for(pe_base::section_list::const_iterator it = sections.begin(); it != sections.end(); ++it, ++current_section)
			{
				//   
				const pe_base::section& s = *it;

				{
					//  
					//      
					packed_section& info
						= reinterpret_cast<packed_section&>(packed_sections_info[current_section * sizeof(packed_section)]);

					// 
					info.characteristics = s.get_characteristics();
					//   
					info.pointer_to_raw_data = s.get_pointer_to_raw_data();
					//  
					info.size_of_raw_data = s.get_size_of_raw_data();
					//   
					info.virtual_address = s.get_virtual_address();
					//  
					info.virtual_size = s.get_virtual_size();

					//   (  8 )
					memset(info.name, 0, sizeof(info.name));
					memcpy(info.name, s.get_name().c_str(), s.get_name().length());
				}

				//  ,   
				if(s.get_raw_data().empty())
					continue;

				//    -     
				//   
				raw_section_data += s.get_raw_data();
			}

			//    ,   !
			if(raw_section_data.empty())
			{
				std::cout << "All sections of PE file are empty!" << std::endl;
				return -1;
			}

			packed_sections_info += raw_section_data;
		}
		

		// 
		pe_base::section new_section;
		// - .rsrc ( )
		new_section.set_name(".rsrc");
		//  , , 
		new_section.readable(true).writeable(true).executable(true);
		//    
		std::string& out_buf = new_section.get_raw_data();


		// "" 
		//      LZO 
		//     
		//  
		//   lzo_align_t  , 
		//    
		//(   LZO)
		boost::scoped_array<lzo_align_t> work_memory(new lzo_align_t[LZO1Z_999_MEM_COMPRESS]);
		
		//  
		lzo_uint src_length = packed_sections_info.size();
		//       
		basic_info.size_of_unpacked_data = src_length;
		
		//  
		//(  )
		lzo_uint out_length = 0;
		
		//    
		//( -     LZO)
		out_buf.resize(src_length + src_length / 16 + 64 + 3);

		//  
		std::cout << "Packing data..." << std::endl;
		if(LZO_E_OK !=
			lzo1z_999_compress(reinterpret_cast<const unsigned char*>(packed_sections_info.data()),
			src_length,
			reinterpret_cast<unsigned char*>(&out_buf[0]),
			&out_length,
			work_memory.get())
			)
		{
			// -  , 
			std::cout << "Error compressing data!" << std::endl;
			return -1;
		}
		
		std::cout << "Packing complete..." << std::endl;

		//      
		basic_info.size_of_packed_data = out_length;
		//      
		//   , 
		//  
		out_buf.resize(out_length);
		//  ,   
		//    
		out_buf =
			//  basic_info
			std::string(reinterpret_cast<const char*>(&basic_info), sizeof(basic_info))
			// 
			+ out_buf;

		//,     
		if(out_buf.size() >= src_length)
		{
			std::cout << "File is incompressible!" << std::endl;
			return -1;
		}


		//   TLS,    
		std::auto_ptr<pe_base::tls_info> tls;
		if(image.has_tls())
		{
			std::cout << "Reading TLS..." << std::endl;
			tls.reset(new pe_base::tls_info(image.get_tls_info()));
		}


		//   ,    
		//  
		pe_base::exported_functions_list exports;
		pe_base::export_info exports_info;
		if(image.has_exports())
		{
			std::cout << "Reading exports..." << std::endl;
			exports = image.get_exported_functions(exports_info);
		}


		//   Image Load Config,    
		std::auto_ptr<pe_base::image_config_info> load_config;
		if(image.has_config())
		{
			std::cout << "Reading Image Load Config..." << std::endl;
			load_config.reset(new pe_base::image_config_info(image.get_image_config()));
		}

		{
			//     
			//  PE-
			const pe_base::section& first_section = image.get_image_sections().front();
			//      ( )
			new_section.set_virtual_address(first_section.get_virtual_address());

			//     
			//  PE-
			const pe_base::section& last_section = image.get_image_sections().back();
			//    
			DWORD total_virtual_size = 
				//   
				last_section.get_virtual_address()
				//    
				+ pe_base::align_up(last_section.get_virtual_size(), image.get_section_alignment())
				//    
				- first_section.get_virtual_address();
			
			//    
			pe_base::resource_directory new_root_dir;

			if(image.has_resources())
			{
				std::cout << "Repacking resources..." << std::endl;

				//    ( )
				pe_base::resource_directory root_dir = image.get_resources();
				//     
				//  
				pe_resource_viewer res(root_dir);
				pe_resource_manager new_res(new_root_dir);

				try
				{
					//    
					//  ,  ID
					pe_resource_viewer::resource_id_list icon_id_list(res.list_resource_ids(pe_resource_viewer::resource_icon_group));
					pe_resource_viewer::resource_name_list icon_name_list(res.list_resource_names(pe_resource_viewer::resource_icon_group));
					//    ,  ,   
					if(!icon_name_list.empty())
					{
						//        (  0)
						//        ,    list_resource_languages
						//        ,    get_icon_by_name (   )
						//      
						new_res.add_icon(
							res.get_icon_by_name(icon_name_list[0]),
							icon_name_list[0],
							res.list_resource_languages(pe_resource_viewer::resource_icon_group, icon_name_list[0]).at(0));
					}
					else if(!icon_id_list.empty()) //    ,     ID
					{
						//        (  0)
						//        ,    list_resource_languages
						//        ,    get_icon_by_id_lang
						//      
						new_res.add_icon(
							res.get_icon_by_id(icon_id_list[0]),
							icon_id_list[0],
							res.list_resource_languages(pe_resource_viewer::resource_icon_group, icon_id_list[0]).at(0));
					}
				}
				catch(const pe_exception&)
				{
					// -   , ,  ,
					//   
				}

				try
				{
					//  ,  ID
					pe_resource_viewer::resource_id_list manifest_id_list(res.list_resource_ids(pe_resource_viewer::resource_manifest));
					if(!manifest_id_list.empty()) //  
					{
						//        (  0)
						//     
						new_res.add_resource(
							res.get_resource_data_by_id(pe_resource_viewer::resource_manifest, manifest_id_list[0]).get_data(),
							pe_resource_viewer::resource_manifest,
							manifest_id_list[0],
							res.list_resource_languages(pe_resource_viewer::resource_manifest, manifest_id_list[0]).at(0)
							);
					}
				}
				catch(const pe_exception&)
				{
					// -   ,
					//   
				}

				try
				{
					//     ,  ID
					pe_resource_viewer::resource_id_list version_info_id_list(res.list_resource_ids(pe_resource_viewer::resource_version));
					if(!version_info_id_list.empty()) //    
					{
						//           (  0)
						//       
						new_res.add_resource(
							res.get_resource_data_by_id(pe_resource_viewer::resource_version, version_info_id_list[0]).get_data(),
							pe_resource_viewer::resource_version,
							version_info_id_list[0],
							res.list_resource_languages(pe_resource_viewer::resource_version, version_info_id_list[0]).at(0)
							);
					}
				}
				catch(const pe_exception&)
				{
					// -   ,
					//   
				}
			}
			
			
			//   PE-
			image.get_image_sections().clear();

			//  ,    
			//,  0x200 -   
			//  PE-
			image.realign_file(0x200);

			//      
			//       
			pe_base::section& added_section = image.add_section(new_section);
			//     
			image.set_section_virtual_size(added_section, total_virtual_size);
			
			
			std::cout << "Creating imports..." << std::endl;

			//    kernel32.dll
			pe_base::import_library kernel32;
			kernel32.set_name("kernel32.dll"); //  
			
			//  
			pe_base::imported_function func;
			func.set_name("LoadLibraryA"); // 
			kernel32.add_import(func); //   
			
			//  
			func.set_name("GetProcAddress");
			kernel32.add_import(func); // 

			//   (RVA)  load_library_a
			//  packed_file_info,     
			//  , ?
			DWORD load_library_address_rva = pe_base::rva_from_section_offset(added_section,
				offsetof(packed_file_info, load_library_a));
			
			//    
			//   (import address table)
			kernel32.set_rva_to_iat(load_library_address_rva);

			//   
			pe_base::imported_functions_list imports;
			//    
			imports.push_back(kernel32);
			
			//  
			pe_base::import_rebuilder_settings settings;
			//Original import address table    ( )
			settings.build_original_iat(false);
			//  IAT    ,
			//  (load_library_address_rva)
			settings.save_iat_and_original_iat_rvas(true, true);
			//      
			settings.set_offset_from_section_start(added_section.get_raw_data().size());
			
			//      ,
			//    
			//   
			if(!new_root_dir.get_entry_list().empty())
				settings.enable_auto_strip_last_section(false);

			// 
			image.rebuild_imports(imports, added_section, settings);

			// ,  ,  
			if(!new_root_dir.get_entry_list().empty())
				image.rebuild_resources(new_root_dir, added_section, added_section.get_raw_data().size());


			//    TLS
			if(tls.get())
			{
				//     
				// 
				//      ,
				//     
				packed_file_info* info = reinterpret_cast<packed_file_info*>(&added_section.get_raw_data()[0]);

				//   
				// TLS
				info->original_tls_index_rva = tls->get_index_rva();

				//    TLS-,   
				//       
				if(!tls->get_tls_callbacks().empty())
					info->original_rva_of_tls_callbacks = tls->get_callbacks_rva();
				
				//     TLS
				//  -        tls_index
				// packed_file_info
				tls->set_index_rva(pe_base::rva_from_section_offset(added_section, offsetof(packed_file_info, tls_index)));
			}
		}


		//    
		//   TLS-
		DWORD first_callback_offset = 0;

		{
			// 
			pe_base::section unpacker_section;
			// - kaimi.ru
			unpacker_section.set_name("kaimi.ru");
			//  ,   
			unpacker_section.readable(true).executable(true).writeable(true);

			{
				std::cout << "Writing unpacker stub, size = " << sizeof(unpacker_data) << " bytes" << std::endl;
				//     
				std::string& unpacker_section_data = unpacker_section.get_raw_data();
				//   
				//     
				//unpacker.h,     main.cpp
				unpacker_section_data = std::string(reinterpret_cast<const char*>(unpacker_data), sizeof(unpacker_data));

				//    
				// 
				*reinterpret_cast<DWORD*>(&unpacker_section_data[original_image_base_offset]) = image.get_image_base_32();
				*reinterpret_cast<DWORD*>(&unpacker_section_data[original_image_base_no_fixup_offset]) = image.get_image_base_32();

				//       ,
				//         
				//    ,   , 
				// packed_file_info
				*reinterpret_cast<DWORD*>(&unpacker_section_data[rva_of_first_section_offset]) = image.get_image_sections().at(0).get_virtual_address();
			}

			//   
			pe_base::section& unpacker_added_section = image.add_section(unpacker_section);

			if(tls.get() || image.has_exports() || image.has_reloc() || load_config.get())
			{
				//     
				//     
				//( ,       
				//    PE)
				unpacker_added_section.get_raw_data().resize(sizeof(unpacker_data));
			}

			
			//    TLS
			if(tls.get())
			{
				std::cout << "Rebuilding TLS..." << std::endl;
				
				//     
				//     
				std::string& data = unpacker_added_section.get_raw_data();

				// ,     IMAGE_TLS_DIRECTORY32
				DWORD directory_pos = data.size();
				//    
				// sizeof(DWORD)   ,  
				//IMAGE_TLS_DIRECTORY32    4-  
				data.resize(data.size() + sizeof(IMAGE_TLS_DIRECTORY32) + sizeof(DWORD));
				
				//  TLS  ...
				if(!tls->get_tls_callbacks().empty())
				{
					//  
					//  TLS-
					// 1    DWORD
					first_callback_offset = data.size();
					data.resize(data.size() + sizeof(DWORD) * (tls->get_tls_callbacks().size() + 1));

					//     (ret 0xC),
					//  
					*reinterpret_cast<DWORD*>(&data[first_callback_offset]) =
						image.rva_to_va_32(pe_base::rva_from_section_offset(unpacker_added_section, empty_tls_callback_offset));
						
					//   
					//  TLS-
					tls->set_callbacks_rva(pe_base::rva_from_section_offset(unpacker_added_section, first_callback_offset));

					//    packed_file_info,  
					//     ,
					//    
					reinterpret_cast<packed_file_info*>(&image.get_image_sections().at(0).get_raw_data()[0])->new_rva_of_tls_callbacks = tls->get_callbacks_rva();
				}
				else
				{
					//  ,     
					tls->set_callbacks_rva(0);
				}
				
				//  ,     
				//   
				tls->clear_tls_callbacks();

				//   
				//     
				tls->set_raw_data_start_rva(pe_base::rva_from_section_offset(unpacker_added_section, data.size()));
				//    
				tls->recalc_raw_data_end_rva();

				// TLS
				// ,       
				//    (  ,  )
				// ,         
				image.rebuild_tls(*tls, unpacker_added_section, directory_pos, false, false, pe_base::tls_data_expand_raw, true, false);

				//    
				//  
				unpacker_added_section.get_raw_data() += tls->get_raw_data();
				//     "kaimi.ru"
				//  SizeOfZeroFill  TLS
				image.set_section_virtual_size(unpacker_added_section, data.size() + tls->get_size_of_zero_fill());

				//,        
				if(!image.has_reloc() && !image.has_exports() && !load_config.get())
					pe_base::strip_nullbytes(unpacker_added_section.get_raw_data());

				//    (  )
				image.prepare_section(unpacker_added_section);
			}


			//    -   
			// ,    
			image.set_ep(image.rva_from_section_offset(unpacker_added_section, 0));
		}
		
		//    
		if(image.has_reloc())
		{
			std::cout << "Creating relocations..." << std::endl;
			
			//      
			pe_base::relocation_table_list reloc_tables;
			pe_base::relocation_table table;
			
			pe_base::section& unpacker_section = image.get_image_sections().at(1);
			
			//    
			//       
			//,        
			// ,    
			table.set_rva(unpacker_section.get_virtual_address());

			//    original_image_base_offset 
			// parameters.h 
			table.add_relocation(pe_base::relocation_entry(original_image_base_offset, IMAGE_REL_BASED_HIGHLOW));


			//    TLS
			if(tls.get())
			{
				//    TLS
				//   
				DWORD tls_directory_offset = image.get_directory_rva(IMAGE_DIRECTORY_ENTRY_TLS)
					- image.section_from_directory(IMAGE_DIRECTORY_ENTRY_TLS).get_virtual_address();

				//    StartAddressOfRawData,
				//EndAddressOfRawData  AddressOfIndex
				//     
				table.add_relocation(pe_base::relocation_entry(static_cast<WORD>(tls_directory_offset + offsetof(IMAGE_TLS_DIRECTORY32, StartAddressOfRawData)), IMAGE_REL_BASED_HIGHLOW));
				table.add_relocation(pe_base::relocation_entry(static_cast<WORD>(tls_directory_offset + offsetof(IMAGE_TLS_DIRECTORY32, EndAddressOfRawData)), IMAGE_REL_BASED_HIGHLOW));
				table.add_relocation(pe_base::relocation_entry(static_cast<WORD>(tls_directory_offset + offsetof(IMAGE_TLS_DIRECTORY32, AddressOfIndex)), IMAGE_REL_BASED_HIGHLOW));

				//  TLS-
				if(first_callback_offset)
				{
					//      AddressOfCallBacks
					//     
					table.add_relocation(pe_base::relocation_entry(static_cast<WORD>(tls_directory_offset + offsetof(IMAGE_TLS_DIRECTORY32, AddressOfCallBacks)), IMAGE_REL_BASED_HIGHLOW));
					table.add_relocation(pe_base::relocation_entry(static_cast<WORD>(first_callback_offset), IMAGE_REL_BASED_HIGHLOW));
				}
			}


			//    
			reloc_tables.push_back(table);

			// ,    
			//   
			image.rebuild_relocations(reloc_tables, unpacker_section, unpacker_section.get_raw_data().size(), true, !image.has_exports() && !load_config.get());
		}


		if(image.has_exports())
		{
			std::cout << "Repacking exports..." << std::endl;

			pe_base::section& unpacker_section = image.get_image_sections().at(1);

			//       "kaimi.ru"
			image.rebuild_exports(exports_info, exports, unpacker_section, unpacker_section.get_raw_data().size(), true, !load_config.get());
		}


		if(load_config.get())
		{
			std::cout << "Repacking load configuration..." << std::endl;

			pe_base::section& unpacker_section = image.get_image_sections().at(1);

			//   Lock-
			load_config->clear_lock_prefix_list();
			load_config->add_lock_prefix_rva(pe_base::rva_from_section_offset(image.get_image_sections().at(0), offsetof(packed_file_info, lock_opcode)));
			
			//         "kaimi.ru"
			//   SE Handler',    Lock-  
			image.rebuild_image_config(*load_config, unpacker_section, unpacker_section.get_raw_data().size(), true, true);
		}


		//    
		//      
		//  ,   
		image.remove_directory(IMAGE_DIRECTORY_ENTRY_BOUND_IMPORT);
		image.remove_directory(IMAGE_DIRECTORY_ENTRY_IAT);
		image.remove_directory(IMAGE_DIRECTORY_ENTRY_SECURITY);
		image.remove_directory(IMAGE_DIRECTORY_ENTRY_DEBUG);
		
		//  ,   
		//  ,    12 ,    
		//    12  
		//image.strip_data_directories(16 - 4); // -   WinXP
		//   ,  - 
		image.strip_stub_overlay();

		
		//  PE-
		//      
		std::string base_file_name(argv[1]);
		std::string dir_name;
		std::string::size_type slash_pos;
		if((slash_pos = base_file_name.find_last_of("/\\")) != std::string::npos)
		{
			dir_name = base_file_name.substr(0, slash_pos + 1); //  
			base_file_name = base_file_name.substr(slash_pos + 1); //  
		}

		//    packed_ + __
		//    ,  
		// ,   
		base_file_name = dir_name + "packed_" + base_file_name;
		// 
		std::ofstream new_pe_file(base_file_name.c_str(), std::ios::out | std::ios::binary | std::ios::trunc);
		if(!new_pe_file)
		{
			//     -  
			std::cout << "Cannot create " << base_file_name << std::endl;
			return -1;
		}

		// PE-
		// DOS-,    NT-
		//(     true)
		//  SizeOfHeaders -     
		image.rebuild_pe(new_pe_file, true, false);

		// ,    
		std::cout << "Packed image was saved to " << base_file_name << std::endl;
	}
	catch(const pe_exception& e)
	{
		//  -     
		//    
		std::cout << "Error: " << e.what() << std::endl;
		return -1;
	}

	return 0;
}
